"""Implementation of SEG-Y standards."""


from segy.schema import SegyStandard
from segy.standards.registry import register_spec
from segy.standards.rev0 import rev0_segy
from segy.standards.rev1 import rev1_segy

register_spec(SegyStandard.REV0, rev0_segy)
register_spec(SegyStandard.REV1, rev1_segy)


__all__ = ["rev0_segy", "rev1_segy", "SegyStandard"]



"""Indexers for SEG-Y files."""

from __future__ import annotations

from abc import ABC
from abc import abstractmethod
from typing import TYPE_CHECKING

import numpy as np
from fsspec.utils import merge_offset_ranges
from pandas import DataFrame

from segy.config import SegyFileSettings
from segy.ibm import ibm2ieee
from segy.schema import Endianness
from segy.schema import ScalarType

if TYPE_CHECKING:
    from pathlib import Path
    from typing import Any

    from fsspec import AbstractFileSystem
    from numpy.typing import NDArray

    from segy.schema import TraceDescriptor
    from segy.schema.base import BaseTypeDescriptor


def trace_ibm2ieee_inplace(trace: NDArray[Any]) -> NDArray[Any]:
    """Convert data of a trace (headers + data) from IBM32 to float32 in place.

    Args:
        trace: A numpy array of type <trace-dtype> containing the trace data.

    Returns:
        A numpy array of type <new-trace-dtype> converted from the input trace data,
        preserving the original header information.

    Note:
        This method converts IBM format trace data to IEEE format inplace, without
        creating a copy of the trace array.
    """
    header_dtype = trace.dtype["header"]
    data_dtype = trace.dtype["data"]

    num_samp = data_dtype.shape
    data_dtype_f32 = np.dtype(("float32", num_samp))

    trace_dtype_fp32 = np.dtype([("header", header_dtype), ("data", data_dtype_f32)])
    trace["data"] = ibm2ieee(trace["data"]).view("uint32")

    return trace.view(trace_dtype_fp32)


def merge_cat_file(
    fs: AbstractFileSystem,
    url: str | Path,
    starts: list[int],
    ends: list[int],
    block_size: int = 8_388_608,
) -> bytearray:
    """Merge sequential byte start/ends and fetch from store.

    Args:
        fs: fsspec FileSystem instance.
        url: Path/URL to file.
        starts: List of start byte locations.
        ends: List of end byte locations.
        block_size: Optional, block size for concurrent downloads.
            Default is 8MiB.

    Returns:
        Bytearray containing all the requested data.
    """
    paths = [url] * len(starts)

    paths, starts, ends = merge_offset_ranges(
        paths,
        starts,
        ends,
        max_block=block_size,
        sort=True,
    )

    buffer_bytes = fs.cat_ranges(
        paths=paths,
        starts=starts,
        ends=ends,
    )

    return bytearray(b"".join(buffer_bytes))


def bounds_check(indices: list[int], max_: int, type_: str) -> None:
    """Check if indices are out of bounds (negative, or more than max).

    Wrapping negative indices is not supported yet. The `type_` argument
    will be used in exceptions to be descriptive.

    Args:
        indices: A list of integer indices.
        max_: The maximum value of the index range.
        type_: The type of indices being checked.

    Raises:
        IndexError: If any of the indices are negative or exceed the maximum value.
    """
    negative_indices = [index for index in indices if index < 0]
    out_of_range_indices = [index for index in indices if index >= max_]

    outliers = negative_indices + out_of_range_indices

    if outliers:
        msg = (
            f"Requested {type_} indices {outliers} are out of bounds. SEG-Y "
            f"file has {max_} traces. Valid indices are "
            f"[0, {max_ - 1})."
        )
        raise IndexError(msg)


class AbstractIndexer(ABC):
    """Abstract class for indexing and fetching structured data from a remote file.

    We calculate byte ranges from indexing of SEG-Y components and use them
    to fetch the data and decode it.

    Args:
        fs: An instance of `fsspec` file-system.
        url: A string representing the URL of the file.
        spec: An instance of BaseTypeDescriptor.
        max_value: An integer representing the maximum value of the index.
        kind: A string representing the kind of index.
        settings: Optional parsing settings.
    """

    def __init__(  # noqa: PLR0913
        self,
        fs: AbstractFileSystem,
        url: str,
        spec: BaseTypeDescriptor,
        max_value: int,
        kind: str,
        settings: SegyFileSettings | None = None,
    ):
        self.fs = fs
        self.url = url
        self.spec = spec
        self.max_value = max_value
        self.kind = kind
        self.settings = SegyFileSettings() if settings is None else settings

    @abstractmethod
    def indices_to_byte_ranges(self, indices: list[int]) -> tuple[list[int], list[int]]:
        """Logic to calculate start/end bytes."""

    @abstractmethod
    def decode(self, buffer: bytearray) -> NDArray[Any]:
        """How to decode the bytes after reading."""

    def __getitem__(self, item: int | list[int] | slice) -> Any:  # noqa: ANN401
        """Operator for integers, lists, and slices with bounds checking."""
        indices = None

        if isinstance(item, int):
            indices = [item]
            bounds_check(indices, self.max_value, self.kind)

        elif isinstance(item, list):
            indices = item
            bounds_check(indices, self.max_value, self.kind)

        elif isinstance(item, slice):
            if item.step == 0:
                msg = "Step of 0 is invalid for slicing."
                raise ValueError(msg)

            start = item.start or 0
            stop = item.stop or self.max_value

            bounds_check([start, stop - 1], self.max_value, self.kind)
            indices = list(range(*item.indices(self.max_value)))

        if len(indices) == 0:
            msg = "Couldn't parse request. Please ensure it is a valid index."
            raise IndexError(msg)

        data = self.fetch(indices)
        return self.post_process(data)

    def post_process(self, data: NDArray[Any]) -> Any:  # noqa: ANN401
        """Optional post-processing. Override in subclass if needed."""
        return data

    def fetch(self, indices: list[int]) -> NDArray[Any]:
        """Fetches and decodes binary data from the given indices.

        Args:
            indices: A list of integers representing the indices.

        Returns:
            An NDArray of any type representing the fetched data.

        Note:
            - This method internally converts the indices to byte ranges using
                the 'indices_to_byte_ranges' method.
            - The byte ranges are used to fetch the corresponding data from the
                file specified by the 'url' parameter.
            - The fetched data is then decoded and squeezed before being returned.
        """
        starts, ends = self.indices_to_byte_ranges(indices)
        buffer = merge_cat_file(self.fs, self.url, starts, ends)
        return self.decode(buffer).squeeze()


class TraceIndexer(AbstractIndexer):
    """Indexer for reading traces (headers + data).

    Inherits from AbstractIndexer. Implements decoding based on trace
    descriptor. It will optionally return the headers as a Pandas
    DataFrame.
    """

    spec: TraceDescriptor

    def indices_to_byte_ranges(self, indices: list[int]) -> tuple[list[int], list[int]]:
        """Convert trace indices to byte ranges."""
        if self.spec.offset is None:
            msg = "Descriptor offset must be specified."
            raise ValueError(msg)

        start_offset = self.spec.offset
        trace_itemsize = self.spec.dtype.itemsize

        starts = [start_offset + i * trace_itemsize for i in indices]
        ends = [start + trace_itemsize for start in starts]

        return starts, ends

    def decode(self, buffer: bytearray) -> NDArray[Any]:
        """Decode whole traces (header + data)."""
        data = np.frombuffer(buffer, dtype=self.spec.dtype)

        if self.settings.ENDIAN == Endianness.BIG:
            data = data.byteswap(inplace=True).newbyteorder()

        if self.spec.data_descriptor.format == ScalarType.IBM32:
            data = trace_ibm2ieee_inplace(data)

        return data

    def post_process(
        self, data: NDArray[Any]
    ) -> NDArray[Any] | dict[str, NDArray[Any] | DataFrame]:
        """Either return struct array or (Header) DataFrame + (Data) Array."""
        if self.settings.USE_PANDAS:
            return {"header": DataFrame(data["header"]), "data": data["data"]}

        return data


class HeaderIndexer(AbstractIndexer):
    """Indexer for reading trace headers only.

    Inherits from AbstractIndexer. Implements decoding based on trace
    descriptor. It will optionally return the headers as a Pandas
    DataFrame.
    """

    spec: TraceDescriptor

    def indices_to_byte_ranges(self, indices: list[int]) -> tuple[list[int], list[int]]:
        """Convert header indices to byte ranges (without trace data)."""
        trace_itemsize = self.spec.dtype.itemsize
        header_itemsize = self.spec.header_descriptor.itemsize

        if self.spec.offset is None:
            msg = "Descriptor offset must be specified."
            raise ValueError(msg)

        start_offset = self.spec.offset

        starts = [start_offset + i * trace_itemsize for i in indices]
        ends = [start + header_itemsize for start in starts]

        return starts, ends

    def decode(self, buffer: bytearray) -> NDArray[Any]:
        """Decode headers only."""
        data = np.frombuffer(buffer, dtype=self.spec.header_descriptor.dtype)

        # TODO(Altay): Handle float/ibm32 etc headers.
        # https://github.com/TGSAI/segy/issues/5
        if self.settings.ENDIAN == Endianness.BIG:
            data = data.byteswap(inplace=True).newbyteorder()

        return data  # noqa: RET504

    def post_process(self, data: NDArray[Any]) -> NDArray[Any] | DataFrame:
        """Either return header as struct array or DataFrame."""
        if self.settings.USE_PANDAS:
            return DataFrame(data)

        # The numpy array breaks downstream logic so for now
        # turning it off and raising a not implemented error.
        msg = "Not using pandas for headers not implemented yet."
        raise NotImplementedError(msg)
        # return bin_hdr.squeeze()


class DataIndexer(AbstractIndexer):
    """Indexer for reading trace data only.

    Inherits from AbstractIndexer. Implements decoding based on trace
    descriptor.
    """

    spec: TraceDescriptor

    def indices_to_byte_ranges(self, indices: list[int]) -> tuple[list[int], list[int]]:
        """Convert data indices to byte ranges (without trace headers)."""
        trace_itemsize = self.spec.dtype.itemsize
        data_itemsize = self.spec.data_descriptor.dtype.itemsize
        header_itemsize = self.spec.header_descriptor.dtype.itemsize

        if self.spec.offset is None:
            msg = "Descriptor offset must be specified."
            raise ValueError(msg)

        start_offset = self.spec.offset + header_itemsize

        starts = [start_offset + i * trace_itemsize for i in indices]
        ends = [start + data_itemsize for start in starts]

        return starts, ends

    def decode(self, buffer: bytearray) -> NDArray[Any]:
        """Decode trace data only."""
        data = np.frombuffer(buffer, dtype=self.spec.data_descriptor.dtype)

        if self.settings.ENDIAN == Endianness.BIG:
            data = data.byteswap(inplace=True).newbyteorder()

        if self.spec.data_descriptor.format == ScalarType.IBM32:
            data = ibm2ieee(data).view("float32")

        return data



"""Descriptor data model implementations for SEG-Y file(s)."""
from __future__ import annotations

from enum import Enum
from typing import TYPE_CHECKING

from pydantic import Field

from segy.schema.base import CamelCaseModel

if TYPE_CHECKING:
    from segy.schema.header import BinaryHeaderDescriptor
    from segy.schema.header import HeaderFieldDescriptor
    from segy.schema.header import TextHeaderDescriptor
    from segy.schema.trace import TraceDataDescriptor
    from segy.schema.trace import TraceDescriptor


class SegyStandard(Enum):
    """Allowed values for SegyStandard in SegyDescriptor."""

    REV0 = 0.0
    REV1 = 1.0
    REV2 = 2.0
    REV21 = 2.1
    CUSTOM = "custom"


class SegyDescriptor(CamelCaseModel):
    """A descriptor class for a SEG-Y file."""

    segy_standard: SegyStandard = Field(
        ..., description="SEG-Y Revision / Standard. Can also be custom."
    )
    text_file_header: TextHeaderDescriptor = Field(
        ..., description="Textual file header descriptor."
    )
    binary_file_header: BinaryHeaderDescriptor = Field(
        ..., description="Binary file header descriptor."
    )
    extended_text_header: TextHeaderDescriptor | None = Field(
        default=None, description="Extended textual header descriptor."
    )
    trace: TraceDescriptor = Field(..., description="Trace header + data descriptor.")

    def customize(  # noqa: PLR0913
        self: SegyDescriptor,
        text_header_spec: TextHeaderDescriptor | None = None,
        binary_header_fields: list[HeaderFieldDescriptor] | None = None,
        extended_text_spec: TextHeaderDescriptor | None = None,
        trace_header_fields: list[HeaderFieldDescriptor] | None = None,
        trace_data_spec: TraceDataDescriptor | None = None,
    ) -> SegyDescriptor:
        """Customize an existing SEG-Y descriptor."""
        new_descr = self.model_copy(deep=True)
        new_descr.segy_standard = SegyStandard.CUSTOM

        if text_header_spec:
            new_descr.text_file_header = text_header_spec

        # Update binary header fields if specified; else will revert to default.
        if binary_header_fields:
            new_descr.binary_file_header.fields = binary_header_fields

        # Update extended text spec if its specified; else will revert to default.
        if extended_text_spec:
            new_descr.extended_text_header = extended_text_spec

        # Update trace header spec if its specified; else will revert to default.
        if trace_header_fields:
            new_descr.trace.header_descriptor.fields = trace_header_fields

        # Update trace data spec if its specified; else will revert to default.
        if trace_data_spec:
            new_descr.trace.data_descriptor = trace_data_spec

        return new_descr


class SegyInfo(CamelCaseModel):
    """Concise and useful information about SEG-Y files."""

    uri: str = Field(..., description="URI of the SEG-Y file.")

    segy_standard: SegyStandard = Field(
        ..., description="SEG-Y Revision / Standard. Can also be custom."
    )

    num_traces: int = Field(..., description="Number of traces.")

    samples_per_trace: int = Field(
        ..., description="Trace length in number of samples."
    )

    sample_interval: int = Field(..., description="Sampling rate from binary header.")

    file_size: float = Field(..., description="File size in GB.")



"""SEG-Y parser configuration."""
from __future__ import annotations

from pydantic import Field
from pydantic_settings import BaseSettings
from pydantic_settings import SettingsConfigDict

from segy.schema import Endianness


class SegyBaseSettings(BaseSettings):
    """Base class for settings."""

    model_config = SettingsConfigDict(extra="ignore", case_sensitive=True)


class SegyFieldSetting(SegyBaseSettings):
    """Setting class to configure a field (key or override)."""

    KEY: str = Field(...)
    VALUE: int | None = Field(...)


class SamplesPerTraceSetting(SegyFieldSetting):
    """Configuration for samples per trace parsing."""

    KEY: str = "samples_per_trace"
    VALUE: int | None = None


class SampleIntervalSetting(SegyFieldSetting):
    """Configuration for samples interval parsing."""

    KEY: str = "sample_interval"
    VALUE: int | None = None


class ExtendedTextHeaderSetting(SegyFieldSetting):
    """Configuration for extended textual headers parsing."""

    KEY: str = "extended_textual_headers"
    VALUE: int | None = None


class SegyBinaryHeaderSettings(SegyBaseSettings):
    """SEG-Y binary header parsing settings."""

    SAMPLES_PER_TRACE: SamplesPerTraceSetting = SamplesPerTraceSetting()
    SAMPLE_INTERVAL: SampleIntervalSetting = SampleIntervalSetting()
    EXTENDED_TEXT_HEADER: ExtendedTextHeaderSetting = ExtendedTextHeaderSetting()


class SegyFileSettings(SegyBaseSettings):
    """SEG-Y file parsing settings."""

    BINARY: SegyBinaryHeaderSettings = SegyBinaryHeaderSettings()
    ENDIAN: Endianness | None = Field(default=Endianness.BIG)
    REVISION: int | None = Field(default=None)

    USE_PANDAS: bool = Field(default=True)

    model_config = SettingsConfigDict(
        env_prefix="SEGY__",
        env_nested_delimiter="__",
    )



"""Compatibility layer for different Python versions."""

import sys

if sys.version_info >= (3, 11):
    from enum import StrEnum
else:
    # Fallback for older Python versions
    from enum import Enum

    class StrEnum(str, Enum):
        """Fallback StrEnum implementation for Python <= 3.10."""

        def __new__(cls, value: str) -> str:  # type: ignore
            """Ensures that enum members are instances of str."""
            member = str.__new__(cls, value)
            member._value_ = value
            return member


__all__ = ["StrEnum"]



"""Low-level floating point conversion operations."""


from __future__ import annotations

from typing import TYPE_CHECKING

import numba as nb
import numpy as np

if TYPE_CHECKING:
    from segy.typing import NDArrayFloat32
    from segy.typing import NDArrayUint32

# IEEE to IBM MASKS ETC
IEEE32_SIGN = np.uint32(0x80000000)
IEEE32_EXPONENT = np.int32(0x7F800000)
IEEE32_FRACTION = np.uint32(0x7FFFFF)

# IBM to IEEE MASKS ETC
BASE2POW24 = np.uint32(0x1000000)
IBM32_EXPONENT = np.uint32(0x7F000000)
IBM32_FRACTION = np.uint32(0xFFFFFF)


@nb.njit(  # type: ignore
    "uint32(float32)",
    cache=True,
    locals={
        "sign": nb.uint32,
        "exponent": nb.int32,
        "exp_remainder": nb.int8,
        "downshift": nb.int8,
        "ibm_mantissa": nb.int32,
    },
)
def ieee2ibm_single(ieee: np.float32) -> np.uint32:
    """IEEE Float to IBM Float conversion.

    Modified from here:
    https://mail.python.org/pipermail/scipy-user/2011-June/029661.html

    Had to do some CPU and memory optimizations + Numba JIT compilation

    Assuming `ieee_array` is little endian and float32. Will convert to float32 if not.
    Returns `ibm_array` as little endian too.

    Byte swapping is up to user after this function.

    Args:
        ieee: Numpy IEEE 32-bit float array.

    Returns:
        IBM 32-bit float converted array with int32 view.
    """
    ieee = np.float32(ieee).view(np.uint32)

    if ieee in [0, 2147483648]:  # 0.0 or np.float32(-0.0).view('uint32')
        return np.uint32(0)

    # Get IEEE's sign and exponent
    sign = ieee & IEEE32_SIGN  # type: ignore
    exponent = ((ieee & IEEE32_EXPONENT) >> 23) - 127  # type: ignore
    # The IBM 7-bit exponent is to the base 16 and the mantissa is presumed to
    # be entirely to the right of the radix point. In contrast, the IEEE
    # exponent is to the base 2 and there is an assumed 1-bit to the left of
    # the radix point.
    # Note: reusing exponent variable, -> it is actually exp16

    # exp16, exp_remainder
    exponent, exp_remainder = divmod(exponent + 1, 4)
    exponent += exp_remainder != 0
    downshift = 4 - exp_remainder if exp_remainder else 0
    exponent = exponent + 64
    # From here down exponent -> ibm_exponent
    exponent = 0 if exponent < 0 else exponent
    exponent = 127 if exponent > 127 else exponent  # noqa: PLR2004
    exponent = exponent << 24
    exponent = exponent if ieee else 0

    # Add the implicit initial 1-bit to the 23-bit IEEE mantissa to get the
    # 24-bit IBM mantissa. Downshift it by the remainder from the exponent's
    # division by 4. It is allowed to have up to 3 leading 0s.
    ibm_mantissa = ((ieee & IEEE32_FRACTION) | 0x800000) >> downshift  # type: ignore

    return sign | exponent | ibm_mantissa  # type: ignore


@nb.njit(  # type: ignore
    "float32(uint32)",
    cache=True,
    locals={
        "sign_bit": nb.boolean,
        "sign": nb.int8,
        "exponent": nb.uint8,
        "mantissa": nb.float32,
        "ieee": nb.float32,
    },
)
def ibm2ieee_single(ibm: np.uint32) -> np.float32:
    """Converts a 32-bit IBM floating point number into 32-bit IEEE format.

    https://en.wikipedia.org/wiki/IBM_hexadecimal_floating-point
    https://en.wikipedia.org/wiki/IEEE_754

    FP Number = (IBM) -1**sign_bit × 0.significand × 16**(exponent−64)

    Args:
        ibm: Value in 32-bit IBM Float in Little-Endian Format.

    Returns:
        Value parsed to 32-bit IEEE Float in Little-Endian Format.
    """
    if ibm & IBM32_FRACTION == 0:
        return np.float32(0)

    sign_bit = ibm >> 31

    exponent = (ibm & IBM32_EXPONENT) >> 24
    mantissa = (ibm & IBM32_FRACTION) / BASE2POW24

    # (1 - 2 * sign_bit) is about 50x faster than (-1)**sign_bit
    sign = 1 - 2 * sign_bit

    # This 16.0 (instead of just 16) is super important.
    # If the base is not a float, it won't work for negative
    # exponents, and fail silently and return zero.
    return sign * mantissa * 16.0 ** (exponent - 64)  # type: ignore


@nb.vectorize("uint32(float32)", cache=True)  # type: ignore
def ieee2ibm(ieee_array: NDArrayFloat32) -> NDArrayUint32:  # pragma: no cover
    """Wrapper for vectorizing IEEE to IBM conversion to arrays."""
    ibm: NDArrayUint32 = ieee2ibm_single(ieee_array)
    return ibm


@nb.vectorize("float32(uint32)", cache=True)  # type: ignore
def ibm2ieee(ibm_array: NDArrayUint32) -> NDArrayFloat32:  # pragma: no cover
    """Wrapper for vectorizing IBM to IEEE conversion to arrays."""
    ieee: NDArrayFloat32 = ibm2ieee_single(ibm_array)
    return ieee



"""Command line interface for segy."""

import json

import typer
from rich import print

from segy import SegyFile
from segy.cli.common import JsonFileOutOption
from segy.cli.common import ListOfFieldNamesOption
from segy.cli.common import ListOfIntegersOption
from segy.cli.common import TextFileOutOption
from segy.cli.common import UriArgument
from segy.cli.common import modify_path
from segy.schema.segy import SegyInfo

BYTE_TO_MB = 1024**2


app = typer.Typer(
    name="dump",
    no_args_is_help=True,
    help="Dump SEG-Y file components to screen or file.",
)


@app.command(rich_help_panel="General Commands")
def info(uri: UriArgument, output: JsonFileOutOption = None) -> None:
    """Get basic information about a SEG-Y file."""
    segy = SegyFile(uri)
    spec = segy.spec

    info = SegyInfo(
        uri=uri,
        segy_standard=spec.segy_standard.value,
        num_traces=segy.num_traces,
        samples_per_trace=segy.binary_header["samples_per_trace"].iloc[0],
        sample_interval=segy.binary_header["sample_interval"].iloc[0],
        file_size=segy.file_size / BYTE_TO_MB,
    )

    info_json = info.model_dump_json(indent=2)

    if output is None:
        print(info_json)

    else:
        output = modify_path(output, suffix="info", default_extension=".json")
        with output.open(mode="w") as f:
            f.write(info_json)
        print(f"Saved SEG-Y info to file {output}")


@app.command(rich_help_panel="Header Commands")
def text_header(uri: UriArgument, output: TextFileOutOption = None) -> None:
    """Print or save the text header of a SEG-Y file."""
    segy = SegyFile(uri)
    text = segy.text_header

    if output is None:
        print(text)

    else:
        output = modify_path(output, suffix="text", default_extension=".txt")
        with output.open(mode="w") as f:
            f.write(text)
        print(f"Saved SEG-Y text header to file {output}")


@app.command(rich_help_panel="Header Commands")
def binary_header(uri: UriArgument, output: JsonFileOutOption = None) -> None:
    """Print or save the binary header of a SEG-Y file."""
    segy = SegyFile(uri)
    bin_header = segy.binary_header
    bin_header_json = bin_header.to_json(orient="records")

    # Extract the first element since we know it's a single row
    parsed_json = json.loads(bin_header_json)
    bin_header_json = json.dumps(parsed_json[0], indent=2)

    if output is None:
        print(parsed_json)

    else:
        output = modify_path(output, suffix="binary", default_extension=".json")
        with output.open(mode="w") as f:
            f.write(bin_header_json)
        print(f"Saved SEG-Y binary header to file {output}")


@app.command(rich_help_panel="Trace Commands")
def trace(
    uri: UriArgument, index: ListOfIntegersOption, output: JsonFileOutOption = None
) -> None:
    """Get one or more traces with headers."""
    segy = SegyFile(uri)

    print(segy.trace[index])

    if output is not None:
        msg = "Trace dump to file is not implemented yet."
        raise NotImplementedError(msg)


@app.command(rich_help_panel="Trace Commands")
def trace_header(
    uri: UriArgument,
    index: ListOfIntegersOption,
    field: ListOfFieldNamesOption,
    output: JsonFileOutOption = None,
) -> None:
    """Get one or more trace's headers (without samples)."""
    segy = SegyFile(uri)
    headers = segy.header[index]

    headers.index[:] = index
    headers.index.name = "trace_index"

    if len(field) > 0:
        headers = headers[field]

    print(headers)

    if output is not None:
        msg = "Trace header dump to file is not implemented yet."
        raise NotImplementedError(msg)


@app.command(rich_help_panel="Trace Commands")
def trace_data(
    uri: UriArgument, index: ListOfIntegersOption, output: JsonFileOutOption = None
) -> None:
    """Get one or more trace's samples (without headers)."""
    segy = SegyFile(uri)

    print(segy.data[index])

    if output is not None:
        msg = "Trace data dump to file is not implemented yet."
        raise NotImplementedError(msg)



"""Classes for managing headers and header groups."""
from __future__ import annotations

from typing import TYPE_CHECKING
from typing import Any

import numpy as np
from pydantic import Field

from segy.compat import StrEnum
from segy.ebcdic import ASCII_TO_EBCDIC
from segy.ebcdic import EBCDIC_TO_ASCII
from segy.schema.base import BaseTypeDescriptor
from segy.schema.data_type import StructuredDataTypeDescriptor
from segy.schema.data_type import StructuredFieldDescriptor

if TYPE_CHECKING:
    from segy.schema.data_type import ScalarType


BinaryHeaderDescriptor = StructuredDataTypeDescriptor
HeaderFieldDescriptor = StructuredFieldDescriptor


class TextHeaderEncoding(StrEnum):
    """Supported textual header encodings."""

    ASCII = "ascii"
    EBCDIC = "ebcdic"


class TextHeaderDescriptor(BaseTypeDescriptor):
    """A descriptor class for SEG-Y textual headers."""

    rows: int = Field(..., description="Number of rows in text header.")
    cols: int = Field(..., description="Number of columns in text header.")
    encoding: TextHeaderEncoding = Field(..., description="String encoding.")
    format: ScalarType = Field(..., description="Type of string.")  # noqa: A003
    offset: int | None = Field(default=None, ge=0, description="Starting byte offset.")

    def __len__(self) -> int:
        """Get length of the textual header (number of characters)."""
        return self.rows * self.cols

    @property
    def dtype(self) -> np.dtype[Any]:
        """Get numpy dtype."""
        return np.dtype((self.format, len(self)))

    def _decode(self, buffer: bytes) -> str:
        """Decode EBCDIC or ASCII bytes into string."""
        if self.encoding == TextHeaderEncoding.EBCDIC:
            buffer_int = np.frombuffer(buffer, dtype=self.dtype)
            buffer = EBCDIC_TO_ASCII[buffer_int].tobytes()

        return buffer.decode("ascii")

    def _encode(self, text_header: str) -> bytes:
        """Encode string to EBCDIC or ASCII bytes."""
        if len(text_header) != len(self):
            msg = "Text length must be equal to rows x cols."
            raise ValueError(msg)

        buffer = text_header.encode("ascii")

        if self.encoding == TextHeaderEncoding.EBCDIC:
            buffer_int = np.frombuffer(buffer, dtype=self.dtype)
            buffer = ASCII_TO_EBCDIC[buffer_int].tobytes()

        return buffer

    def _wrap(self, string: str) -> str:
        """Wrap text header string to be multi-line with 80 character columns."""
        if len(string) != len(self):
            msg = "rows x cols must be equal wrapped text length."
            raise ValueError(msg)

        rows = []
        for row_idx in range(self.rows):
            start = row_idx * self.cols
            stop = start + self.cols
            rows.append(string[start:stop])

        return "\n".join(rows)

    @staticmethod
    def _unwrap(text_header: str) -> str:
        """Unwrap a multi-line string to a single line."""
        return text_header.replace("\n", "")



"""Implements a registry for various SEG-Y standards."""
from __future__ import annotations

from typing import TYPE_CHECKING

from segy.schema.segy import SegyDescriptor

if TYPE_CHECKING:
    from segy.schema.segy import SegyStandard

registry = {}


def register_spec(spec_type: SegyStandard, spec: SegyDescriptor) -> None:
    """Register a SEG-Y standard with its descriptor."""
    is_segy_descriptor = isinstance(spec, SegyDescriptor)

    if not is_segy_descriptor:
        msg = "spec_cls must be a subclass of SegyDescriptor."
        raise ValueError(msg)

    registry[spec_type] = spec


def get_spec(spec_type: SegyStandard) -> SegyDescriptor:
    """Get a registered SEG-Y standard's descriptor."""
    spec = registry.get(spec_type)

    if spec is None:
        msg = (
            f"Unknown or unsupported SEG-Y spec: {spec_type}. If you "
            f"would like to use {spec_type}, please register it with "
            f"the `SegySpecFactory` using its `register_spec` method."
        )
        raise NotImplementedError(msg)

    return spec.model_copy(deep=True)



"""Data models for manipulating all kinds of SEG-Y files."""


from segy.schema.data_type import Endianness
from segy.schema.data_type import ScalarType
from segy.schema.header import BinaryHeaderDescriptor
from segy.schema.header import HeaderFieldDescriptor
from segy.schema.header import TextHeaderDescriptor
from segy.schema.header import TextHeaderEncoding
from segy.schema.segy import SegyDescriptor
from segy.schema.segy import SegyStandard
from segy.schema.trace import TraceDataDescriptor
from segy.schema.trace import TraceDescriptor
from segy.schema.trace import TraceHeaderDescriptor

__all__ = [
    "Endianness",
    "ScalarType",
    "TextHeaderDescriptor",
    "TextHeaderEncoding",
    "HeaderFieldDescriptor",
    "BinaryHeaderDescriptor",
    "TraceHeaderDescriptor",
    "TraceDataDescriptor",
    "TraceDescriptor",
    "SegyDescriptor",
    "SegyStandard",
]



"""SEG-Y Revision 1 Specification."""

from segy.schema import BinaryHeaderDescriptor
from segy.schema import Endianness
from segy.schema import HeaderFieldDescriptor
from segy.schema import ScalarType
from segy.schema import SegyDescriptor
from segy.schema import SegyStandard
from segy.schema import TextHeaderDescriptor
from segy.schema import TextHeaderEncoding
from segy.schema import TraceDataDescriptor
from segy.schema import TraceDescriptor
from segy.schema import TraceHeaderDescriptor
from segy.standards.rev0 import BINARY_FILE_HEADER_FIELDS_REV0
from segy.standards.rev0 import TRACE_HEADER_FIELDS_REV0

BINARY_FILE_HEADER_FIELDS_REV1 = BINARY_FILE_HEADER_FIELDS_REV0 + [
    HeaderFieldDescriptor(
        name="seg_y_revision",
        offset=300,
        format=ScalarType.INT16,
        endianness=Endianness.BIG,
        description="SEG Y Format Revision Number",
    ),
    HeaderFieldDescriptor(
        name="fixed_length_trace_flag",
        offset=302,
        format=ScalarType.INT16,
        endianness=Endianness.BIG,
        description="Fixed Length Trace Flag",
    ),
    HeaderFieldDescriptor(
        name="extended_textual_headers",
        offset=304,
        format=ScalarType.INT16,
        endianness=Endianness.BIG,
        description="Number of 3200-byte, Extended Textual File Header Records Following the Binary Header",  # noqa: E501
    ),
    HeaderFieldDescriptor(
        name="additional_trace_headers",
        offset=306,
        format=ScalarType.INT16,
        endianness=Endianness.BIG,
        description="Maximum Number of Additional Trace Headers",
    ),
]


TRACE_HEADER_FIELDS_REV1 = TRACE_HEADER_FIELDS_REV0 + [
    HeaderFieldDescriptor(
        name="x_coordinate",
        offset=180,
        format=ScalarType.INT32,
        endianness=Endianness.BIG,
        description="X coordinate of ensemble (CDP) position",
    ),
    HeaderFieldDescriptor(
        name="y_coordinate",
        offset=184,
        format=ScalarType.INT32,
        endianness=Endianness.BIG,
        description="Y coordinate of ensemble (CDP) position",
    ),
    HeaderFieldDescriptor(
        name="inline_no",
        offset=188,
        format=ScalarType.INT32,
        endianness=Endianness.BIG,
        description="Inline number",
    ),
    HeaderFieldDescriptor(
        name="crossline_no",
        offset=192,
        format=ScalarType.INT32,
        endianness=Endianness.BIG,
        description="Crossline number",
    ),
    HeaderFieldDescriptor(
        name="shotpoint_no",
        offset=196,
        format=ScalarType.INT32,
        endianness=Endianness.BIG,
        description="Shotpoint number",
    ),
    HeaderFieldDescriptor(
        name="scalar_apply_shotpoint",
        offset=200,
        format=ScalarType.INT16,
        endianness=Endianness.BIG,
        description="Scalar to be applied to the shotpoint number",
    ),
    HeaderFieldDescriptor(
        name="trace_value_measurement_unit",
        offset=202,
        format=ScalarType.INT16,
        endianness=Endianness.BIG,
        description="Trace value measurement unit",
    ),
    HeaderFieldDescriptor(
        name="transduction_constant_mantissa",
        offset=204,
        format=ScalarType.INT32,
        endianness=Endianness.BIG,
        description="Transduction Constant Mantissa",
    ),
    HeaderFieldDescriptor(
        name="transduction_constant_exponent",
        offset=208,
        format=ScalarType.INT16,
        endianness=Endianness.BIG,
        description="Transduction Constant Exponent",
    ),
    HeaderFieldDescriptor(
        name="transduction_units",
        offset=210,
        format=ScalarType.INT16,
        endianness=Endianness.BIG,
        description="Transduction Units",
    ),
    HeaderFieldDescriptor(
        name="device_trace_id",
        offset=212,
        format=ScalarType.INT16,
        endianness=Endianness.BIG,
        description="Device/Trace Identifier",
    ),
    HeaderFieldDescriptor(
        name="times_scalar",
        offset=214,
        format=ScalarType.INT16,
        endianness=Endianness.BIG,
        description="Scalar to be applied to times",
    ),
    HeaderFieldDescriptor(
        name="source_type_orientation",
        offset=216,
        format=ScalarType.INT16,
        endianness=Endianness.BIG,
        description="Source Type/Orientation",
    ),
    HeaderFieldDescriptor(
        name="source_energy_direction_mantissa",
        offset=218,
        format=ScalarType.INT32,
        endianness=Endianness.BIG,
        description="Source Energy Direction with respect to vertical [Mantissa]",
    ),
    HeaderFieldDescriptor(
        name="source_energy_direction_exponent",
        offset=222,
        format=ScalarType.INT16,
        endianness=Endianness.BIG,
        description="Source Energy Direction with respect to vertical [Exponent]",
    ),
    HeaderFieldDescriptor(
        name="source_measurement_mantissa",
        offset=224,
        format=ScalarType.INT32,
        endianness=Endianness.BIG,
        description="Source Measurement Mantissa",
    ),
    HeaderFieldDescriptor(
        name="source_measurement_exponent",
        offset=228,
        format=ScalarType.INT16,
        endianness=Endianness.BIG,
        description="Source Measurement Exponent",
    ),
    HeaderFieldDescriptor(
        name="source_measurement_unit",
        offset=230,
        format=ScalarType.INT16,
        endianness=Endianness.BIG,
        description="Source Measurement Unit",
    ),
]


rev1_textual_file_header = TextHeaderDescriptor(
    rows=40,
    cols=80,
    offset=0,
    encoding=TextHeaderEncoding.EBCDIC,
    format=ScalarType.UINT8,  # noqa: A003
)


rev1_binary_file_header = BinaryHeaderDescriptor(
    fields=BINARY_FILE_HEADER_FIELDS_REV1,
    item_size=400,
    offset=3200,
)


rev1_extended_text_header = TextHeaderDescriptor(
    rows=40,
    cols=80,
    offset=3600,
    encoding=TextHeaderEncoding.EBCDIC,
    format=ScalarType.UINT8,  # noqa: A003
)


rev1_trace_header = TraceHeaderDescriptor(
    fields=TRACE_HEADER_FIELDS_REV1,
    item_size=240,
)


rev1_trace_data = TraceDataDescriptor(
    format=ScalarType.IBM32,  # noqa: A003
    endianness=Endianness.BIG,
)


rev1_trace = TraceDescriptor(
    header_descriptor=rev1_trace_header,
    data_descriptor=rev1_trace_data,
)


rev1_segy = SegyDescriptor(
    segy_standard=SegyStandard.REV1,
    text_file_header=rev1_textual_file_header,
    binary_file_header=rev1_binary_file_header,
    extended_text_header=rev1_extended_text_header,
    trace=rev1_trace,
)



"""SEG-Y file in-memory constructors."""
from __future__ import annotations

from functools import cached_property
from typing import TYPE_CHECKING

import numpy as np
from fsspec.core import url_to_fs
from pandas import DataFrame

from segy.config import SegyFileSettings
from segy.indexing import DataIndexer
from segy.indexing import HeaderIndexer
from segy.indexing import TraceIndexer
from segy.schema import Endianness
from segy.schema import SegyStandard
from segy.standards.registry import get_spec
from segy.standards.rev1 import rev1_binary_file_header
from segy.standards.rev1 import rev1_extended_text_header

if TYPE_CHECKING:
    from typing import Any

    from fsspec import AbstractFileSystem

    from segy.indexing import AbstractIndexer
    from segy.schema import SegyDescriptor


class SegyFile:
    """A SEG-Y file opener with fancy features."""

    fs: AbstractFileSystem
    url: str

    def __init__(
        self,
        url: str,
        spec: SegyDescriptor | None = None,
        settings: SegyFileSettings | None = None,
        storage_options: dict[str, Any] | None = None,
    ):
        self.settings = SegyFileSettings() if settings is None else settings

        if storage_options is None:
            storage_options = {}

        self.fs, self.url = url_to_fs(url, **storage_options)

        self.spec = self._infer_standard() if spec is None else spec

        self._info = self.fs.info(self.url)
        self._parse_binary_header()

    @classmethod
    def from_spec(
        cls: type[SegyFile],
        url: str,
        spec: SegyDescriptor,
        storage_options: dict[str, Any] | None = None,
    ) -> SegyFile:
        """Open a SEG-Y file based on custom spec."""
        return cls(url=url, spec=spec, storage_options=storage_options)

    @property
    def file_size(self) -> int:
        """Return file size in bytes."""
        size: int = self._info["size"]
        return size

    @property
    def num_traces(self) -> int:
        """Return number of traces in file based on size and spec."""
        file_textual_hdr_size = self.spec.text_file_header.itemsize
        file_bin_hdr_size = self.spec.binary_file_header.itemsize
        trace_size = self.spec.trace.itemsize

        num_ext_text = 0
        # All but Rev0 can have extended text headers
        if self.spec.segy_standard != SegyStandard.REV0:
            header_key = self.settings.BINARY.EXTENDED_TEXT_HEADER.KEY

            num_ext_text = 0
            if header_key in self.binary_header:
                # Use df.at method to get a single value and not return a series
                num_ext_text = self.binary_header.at[0, header_key]

        # Overriding from settings
        elif self.settings.BINARY.EXTENDED_TEXT_HEADER.VALUE is not None:
            num_ext_text = self.settings.BINARY.EXTENDED_TEXT_HEADER.VALUE

        file_metadata_size = file_textual_hdr_size + file_bin_hdr_size

        if num_ext_text > 0:
            self.spec.extended_text_header = rev1_extended_text_header

            ext_text_size = self.spec.extended_text_header.itemsize * int(num_ext_text)
            file_metadata_size = file_metadata_size + ext_text_size

        return (self.file_size - file_metadata_size) // trace_size

    @cached_property
    def text_header(self) -> str:
        """Return textual file header."""
        text_hdr_desc = self.spec.text_file_header

        buffer = self.fs.read_block(
            fn=self.url,
            offset=text_hdr_desc.offset,
            length=text_hdr_desc.itemsize,
        )

        text_header = text_hdr_desc._decode(buffer)
        return text_hdr_desc._wrap(text_header)

    def _infer_standard(self) -> SegyDescriptor:
        if self.settings.REVISION is None:
            buffer = self.fs.read_block(
                fn=self.url,
                offset=rev1_binary_file_header.offset,
                length=rev1_binary_file_header.item_size,
            )
            revision = np.frombuffer(buffer, offset=300, dtype=">i2", count=1).item()

            # Per SEG-Y standard, there is a Q-point between the bytes. Dividing
            # by 2^8 to get the floating-point value of the revision.
            revision = revision / 256.0

        else:
            revision = self.settings.REVISION

        standard = SegyStandard(revision)
        return get_spec(standard)

    @cached_property
    def binary_header(self) -> DataFrame:
        """Read binary header from store, based on spec."""
        buffer = bytearray(
            self.fs.read_block(
                fn=self.url,
                offset=self.spec.binary_file_header.offset,
                length=self.spec.binary_file_header.itemsize,
            )
        )

        bin_hdr = np.frombuffer(buffer, dtype=self.spec.binary_file_header.dtype)

        if self.settings.ENDIAN == Endianness.BIG:
            bin_hdr = bin_hdr.byteswap(inplace=True).newbyteorder()

        if self.settings.USE_PANDAS:
            return DataFrame(bin_hdr.reshape(1))

        # The numpy array breaks downstream logic so for now
        # turning it off and raising a not implemented error.
        msg = "Not using pandas for headers not implemented yet."
        raise NotImplementedError(msg)
        # return bin_hdr.squeeze()

    def _parse_binary_header(self) -> None:
        """Parse the binary header and apply some rules."""
        # Extract number of samples and extended text headers.
        if self.settings.BINARY.SAMPLES_PER_TRACE.VALUE is None:
            header_key = self.settings.BINARY.SAMPLES_PER_TRACE.KEY
            # Use df.at method to get a single value and not return a series
            samples_per_trace = self.binary_header.at[0, header_key]
        else:
            samples_per_trace = self.settings.BINARY.SAMPLES_PER_TRACE.VALUE

        # Calculate sizes for dynamic file metadata
        text_hdr_size = self.spec.text_file_header.itemsize
        bin_hdr_size = self.spec.binary_file_header.itemsize

        # Update trace start offset and sample length
        self.spec.trace.data_descriptor.samples = int(samples_per_trace)
        self.spec.trace.offset = text_hdr_size + bin_hdr_size

        num_ext_text = 0
        # All but Rev0 can have extended text headers
        if self.spec.segy_standard != SegyStandard.REV0:
            header_key = self.settings.BINARY.EXTENDED_TEXT_HEADER.KEY

            num_ext_text = 0
            if header_key in self.binary_header:
                # Use df.at method to get a single value and not return a series
                num_ext_text = self.binary_header.at[0, header_key]

        # Overriding from settings
        elif self.settings.BINARY.EXTENDED_TEXT_HEADER.VALUE is not None:
            num_ext_text = self.settings.BINARY.EXTENDED_TEXT_HEADER.VALUE

        if num_ext_text > 0:
            self.spec.extended_text_header = rev1_extended_text_header

            ext_text_size = self.spec.extended_text_header.itemsize * int(num_ext_text)
            self.spec.trace.offset = self.spec.trace.offset + ext_text_size

    @property
    def data(self) -> AbstractIndexer:
        """Way to access the file to fetch trace data only."""
        return DataIndexer(
            self.fs,
            self.url,
            self.spec.trace,
            self.num_traces,
            kind="data",
            settings=self.settings,
        )

    @property
    def header(self) -> AbstractIndexer:
        """Way to access the file to fetch trace headers only."""
        return HeaderIndexer(
            self.fs,
            self.url,
            self.spec.trace,
            self.num_traces,
            kind="header",
            settings=self.settings,
        )

    @property
    def trace(self) -> AbstractIndexer:
        """Way to access the file to fetch full traces (headers + data)."""
        return TraceIndexer(
            self.fs,
            self.url,
            self.spec.trace,
            self.num_traces,
            kind="trace",
            settings=self.settings,
        )



"""Descriptor data model implementations for traces."""
from __future__ import annotations

from typing import TYPE_CHECKING
from typing import Any

import numpy as np
from pydantic import Field

from segy.schema.base import BaseTypeDescriptor
from segy.schema.data_type import Endianness
from segy.schema.data_type import StructuredDataTypeDescriptor

if TYPE_CHECKING:
    from segy.schema.data_type import ScalarType


TraceHeaderDescriptor = StructuredDataTypeDescriptor


class TraceDataDescriptor(BaseTypeDescriptor):
    """A descriptor class for a Trace Data (samples)."""

    format: ScalarType = Field(..., description="Format of trace samples.")  # noqa: A003
    endianness: Endianness = Field(
        default=Endianness.BIG, description="Endianness of trace samples."
    )
    samples: int | None = Field(
        default=None,
        description=(
            "Number of samples in trace. It can be variable, "
            "then it must be read from each trace header."
        ),
    )

    @property
    def dtype(self) -> np.dtype[Any]:
        """Get numpy dtype."""
        format_char = self.format.char
        dtype_str = "".join([self.endianness.symbol, str(self.samples), format_char])
        return np.dtype(dtype_str)


class TraceDescriptor(BaseTypeDescriptor):
    """A descriptor class for a Trace (Header + Data)."""

    header_descriptor: TraceHeaderDescriptor = Field(
        ..., description="Trace header descriptor."
    )
    extended_header_descriptor: TraceHeaderDescriptor | None = Field(
        default=None, description="Extended trace header descriptor."
    )
    data_descriptor: TraceDataDescriptor = Field(
        ..., description="Trace data descriptor."
    )
    offset: int | None = Field(
        default=None, description="Starting offset of the trace."
    )

    @property
    def dtype(self) -> np.dtype[Any]:
        """Get numpy dtype."""
        header_dtype = self.header_descriptor.dtype
        data_dtype = self.data_descriptor.dtype

        return np.dtype([("header", header_dtype), ("data", data_dtype)])



"""SEG-Y Revision 0 Specification."""

from segy.schema import BinaryHeaderDescriptor
from segy.schema import Endianness
from segy.schema import HeaderFieldDescriptor
from segy.schema import ScalarType
from segy.schema import SegyDescriptor
from segy.schema import SegyStandard
from segy.schema import TextHeaderDescriptor
from segy.schema import TextHeaderEncoding
from segy.schema import TraceDataDescriptor
from segy.schema import TraceDescriptor
from segy.schema import TraceHeaderDescriptor

BINARY_FILE_HEADER_FIELDS_REV0 = [
    HeaderFieldDescriptor(
        name="job_id",
        offset=0,
        format=ScalarType.INT32,
        endianness=Endianness.BIG,
        description="Job Identification Number",
    ),
    HeaderFieldDescriptor(
        name="line_no",
        offset=4,
        format=ScalarType.INT32,
        endianness=Endianness.BIG,
        description="Line Number",
    ),
    HeaderFieldDescriptor(
        name="reel_no",
        offset=8,
        format=ScalarType.INT32,
        endianness=Endianness.BIG,
        description="Reel Number",
    ),
    HeaderFieldDescriptor(
        name="data_traces_ensemble",
        offset=12,
        format=ScalarType.INT16,
        endianness=Endianness.BIG,
        description="Number of Data Traces per Ensemble",
    ),
    HeaderFieldDescriptor(
        name="aux_traces_ensemble",
        offset=14,
        format=ScalarType.INT16,
        endianness=Endianness.BIG,
        description="Number of Auxiliary Traces per Ensemble",
    ),
    HeaderFieldDescriptor(
        name="sample_interval",
        offset=16,
        format=ScalarType.INT16,
        endianness=Endianness.BIG,
        description="Sample Interval",
    ),
    HeaderFieldDescriptor(
        name="sample_interval_orig",
        offset=18,
        format=ScalarType.INT16,
        endianness=Endianness.BIG,
        description="Sample Interval of Original Field Recording",
    ),
    HeaderFieldDescriptor(
        name="samples_per_trace",
        offset=20,
        format=ScalarType.INT16,
        endianness=Endianness.BIG,
        description="Number of Samples per Data Trace",
    ),
    HeaderFieldDescriptor(
        name="samples_per_trace_orig",
        offset=22,
        format=ScalarType.INT16,
        endianness=Endianness.BIG,
        description="Number of Samples per Data Trace for Original Field Recording",
    ),
    HeaderFieldDescriptor(
        name="data_sample_format",
        offset=24,
        format=ScalarType.INT16,
        endianness=Endianness.BIG,
        description="Data Sample Format Code",
    ),
    HeaderFieldDescriptor(
        name="ensemble_fold",
        offset=26,
        format=ScalarType.INT16,
        endianness=Endianness.BIG,
        description="Ensemble Fold",
    ),
    HeaderFieldDescriptor(
        name="trace_sorting",
        offset=28,
        format=ScalarType.INT16,
        endianness=Endianness.BIG,
        description="Trace Sorting Code",
    ),
    HeaderFieldDescriptor(
        name="vertical_sum",
        offset=30,
        format=ScalarType.INT16,
        endianness=Endianness.BIG,
        description="Vertical Sum Code",
    ),
    HeaderFieldDescriptor(
        name="sweep_freq_start",
        offset=32,
        format=ScalarType.INT16,
        endianness=Endianness.BIG,
        description="Sweep Frequency at Start",
    ),
    HeaderFieldDescriptor(
        name="sweep_freq_end",
        offset=34,
        format=ScalarType.INT16,
        endianness=Endianness.BIG,
        description="Sweep Frequency at End",
    ),
    HeaderFieldDescriptor(
        name="sweep_length",
        offset=36,
        format=ScalarType.INT16,
        endianness=Endianness.BIG,
        description="Sweep Length",
    ),
    HeaderFieldDescriptor(
        name="sweep_type",
        offset=38,
        format=ScalarType.INT16,
        endianness=Endianness.BIG,
        description="Sweep Type Code",
    ),
    HeaderFieldDescriptor(
        name="sweep_trace_no",
        offset=40,
        format=ScalarType.INT16,
        endianness=Endianness.BIG,
        description="Trace Number of Sweep Channel",
    ),
    HeaderFieldDescriptor(
        name="sweep_taper_start",
        offset=42,
        format=ScalarType.INT16,
        endianness=Endianness.BIG,
        description="Sweep Trace Taper Length at Start",
    ),
    HeaderFieldDescriptor(
        name="sweep_taper_end",
        offset=44,
        format=ScalarType.INT16,
        endianness=Endianness.BIG,
        description="Sweep Trace Taper Length at End",
    ),
    HeaderFieldDescriptor(
        name="taper_type",
        offset=46,
        format=ScalarType.INT16,
        endianness=Endianness.BIG,
        description="Taper Type",
    ),
    HeaderFieldDescriptor(
        name="correlated_traces",
        offset=48,
        format=ScalarType.INT16,
        endianness=Endianness.BIG,
        description="Correlated Data Traces",
    ),
    HeaderFieldDescriptor(
        name="binary_gain",
        offset=50,
        format=ScalarType.INT16,
        endianness=Endianness.BIG,
        description="Binary Gain Recovered",
    ),
    HeaderFieldDescriptor(
        name="amp_recovery_method",
        offset=52,
        format=ScalarType.INT16,
        endianness=Endianness.BIG,
        description="Amplitude Recovery Method",
    ),
    HeaderFieldDescriptor(
        name="measurement_system",
        offset=54,
        format=ScalarType.INT16,
        endianness=Endianness.BIG,
        description="Measurement System",
    ),
    HeaderFieldDescriptor(
        name="impulse_signal_polarity",
        offset=56,
        format=ScalarType.INT16,
        endianness=Endianness.BIG,
        description="Impulse Signal Polarity",
    ),
    HeaderFieldDescriptor(
        name="vibratory_polarity",
        offset=58,
        format=ScalarType.INT16,
        endianness=Endianness.BIG,
        description="Vibratory Polarity Code",
    ),
]


TRACE_HEADER_FIELDS_REV0 = [
    HeaderFieldDescriptor(
        name="trace_seq_line",
        offset=0,
        format=ScalarType.INT32,
        endianness=Endianness.BIG,
        description="Trace Sequence Number within Line",
    ),
    HeaderFieldDescriptor(
        name="trace_seq_file",
        offset=4,
        format=ScalarType.INT32,
        endianness=Endianness.BIG,
        description="Trace Sequence Number within File",
    ),
    HeaderFieldDescriptor(
        name="field_rec_no",
        offset=8,
        format=ScalarType.INT32,
        endianness=Endianness.BIG,
        description="Original Field Record Number",
    ),
    HeaderFieldDescriptor(
        name="trace_no_field_rec",
        offset=12,
        format=ScalarType.INT32,
        endianness=Endianness.BIG,
        description="Trace Number within the Field Record",
    ),
    HeaderFieldDescriptor(
        name="energy_src_pt",
        offset=16,
        format=ScalarType.INT32,
        endianness=Endianness.BIG,
        description="Energy Source Point Number",
    ),
    HeaderFieldDescriptor(
        name="cdp_ens_no",
        offset=20,
        format=ScalarType.INT32,
        endianness=Endianness.BIG,
        description="Ensemble Number (CDP, CMP, etc.)",
    ),
    HeaderFieldDescriptor(
        name="trace_no_ens",
        offset=24,
        format=ScalarType.INT32,
        endianness=Endianness.BIG,
        description="Trace Number within the Ensemble",
    ),
    HeaderFieldDescriptor(
        name="trace_id",
        offset=28,
        format=ScalarType.INT16,
        endianness=Endianness.BIG,
        description="Trace Identification Code",
    ),
    HeaderFieldDescriptor(
        name="vert_sum",
        offset=30,
        format=ScalarType.INT16,
        endianness=Endianness.BIG,
        description="Number of Vertically Stacked Traces",
    ),
    HeaderFieldDescriptor(
        name="horiz_stack",
        offset=32,
        format=ScalarType.INT16,
        endianness=Endianness.BIG,
        description="Number of Horizontally Stacked Traces",
    ),
    HeaderFieldDescriptor(
        name="data_use",
        offset=34,
        format=ScalarType.INT16,
        endianness=Endianness.BIG,
        description="Data Use",
    ),
    HeaderFieldDescriptor(
        name="dist_src_to_rec",
        offset=36,
        format=ScalarType.INT32,
        endianness=Endianness.BIG,
        description="Distance from Source Point to Receiver Group",
    ),
    HeaderFieldDescriptor(
        name="rec_elev",
        offset=40,
        format=ScalarType.INT32,
        endianness=Endianness.BIG,
        description="Receiver Group Elevation",
    ),
    HeaderFieldDescriptor(
        name="src_elev",
        offset=44,
        format=ScalarType.INT32,
        endianness=Endianness.BIG,
        description="Source Elevation",
    ),
    HeaderFieldDescriptor(
        name="src_depth",
        offset=48,
        format=ScalarType.INT32,
        endianness=Endianness.BIG,
        description="Source Depth",
    ),
    HeaderFieldDescriptor(
        name="datum_elev_rec",
        offset=52,
        format=ScalarType.INT32,
        endianness=Endianness.BIG,
        description="Datum Elevation at Receiver Group",
    ),
    HeaderFieldDescriptor(
        name="datum_elev_src",
        offset=56,
        format=ScalarType.INT32,
        endianness=Endianness.BIG,
        description="Datum Elevation at Source",
    ),
    HeaderFieldDescriptor(
        name="water_depth_src",
        offset=60,
        format=ScalarType.INT32,
        endianness=Endianness.BIG,
        description="Water Depth at Source",
    ),
    HeaderFieldDescriptor(
        name="water_depth_rec",
        offset=64,
        format=ScalarType.INT32,
        endianness=Endianness.BIG,
        description="Water Depth at Receiver Group",
    ),
    HeaderFieldDescriptor(
        name="scalar_apply_elev",
        offset=68,
        format=ScalarType.INT16,
        endianness=Endianness.BIG,
        description="Scalar to be applied to all elevations and depths",
    ),
    HeaderFieldDescriptor(
        name="scalar_apply_coords",
        offset=70,
        format=ScalarType.INT16,
        endianness=Endianness.BIG,
        description="Scalar to be applied to all coordinates",
    ),
    HeaderFieldDescriptor(
        name="src_x",
        offset=72,
        format=ScalarType.INT32,
        endianness=Endianness.BIG,
        description="Source X coordinate",
    ),
    HeaderFieldDescriptor(
        name="src_y",
        offset=76,
        format=ScalarType.INT32,
        endianness=Endianness.BIG,
        description="Source Y coordinate",
    ),
    HeaderFieldDescriptor(
        name="rec_x",
        offset=80,
        format=ScalarType.INT32,
        endianness=Endianness.BIG,
        description="Receiver X coordinate",
    ),
    HeaderFieldDescriptor(
        name="rec_y",
        offset=84,
        format=ScalarType.INT32,
        endianness=Endianness.BIG,
        description="Receiver Y coordinate",
    ),
    HeaderFieldDescriptor(
        name="coord_units",
        offset=88,
        format=ScalarType.INT16,
        endianness=Endianness.BIG,
        description="Coordinate units",
    ),
    HeaderFieldDescriptor(
        name="weathering_vel",
        offset=90,
        format=ScalarType.INT16,
        endianness=Endianness.BIG,
        description="Weathering Velocity",
    ),
    HeaderFieldDescriptor(
        name="subweathering_vel",
        offset=92,
        format=ScalarType.INT16,
        endianness=Endianness.BIG,
        description="Subweathering Velocity",
    ),
    HeaderFieldDescriptor(
        name="uphole_time_src",
        offset=94,
        format=ScalarType.INT16,
        endianness=Endianness.BIG,
        description="Uphole Time at Source",
    ),
    HeaderFieldDescriptor(
        name="uphole_time_rec",
        offset=96,
        format=ScalarType.INT16,
        endianness=Endianness.BIG,
        description="Uphole Time at Receiver",
    ),
    HeaderFieldDescriptor(
        name="src_static_corr",
        offset=98,
        format=ScalarType.INT16,
        endianness=Endianness.BIG,
        description="Source Static Correction",
    ),
    HeaderFieldDescriptor(
        name="rec_static_corr",
        offset=100,
        format=ScalarType.INT16,
        endianness=Endianness.BIG,
        description="Receiver Static Correction",
    ),
    HeaderFieldDescriptor(
        name="total_static",
        offset=102,
        format=ScalarType.INT16,
        endianness=Endianness.BIG,
        description="Total Static Applied",
    ),
    HeaderFieldDescriptor(
        name="lag_time_a",
        offset=104,
        format=ScalarType.INT16,
        endianness=Endianness.BIG,
        description="Lag Time A",
    ),
    HeaderFieldDescriptor(
        name="lag_time_b",
        offset=106,
        format=ScalarType.INT16,
        endianness=Endianness.BIG,
        description="Lag Time B",
    ),
    HeaderFieldDescriptor(
        name="delay_rec_time",
        offset=108,
        format=ScalarType.INT16,
        endianness=Endianness.BIG,
        description="Delay Recording Time",
    ),
    HeaderFieldDescriptor(
        name="mute_start",
        offset=110,
        format=ScalarType.INT16,
        endianness=Endianness.BIG,
        description="Start Time of Mute",
    ),
    HeaderFieldDescriptor(
        name="mute_end",
        offset=112,
        format=ScalarType.INT16,
        endianness=Endianness.BIG,
        description="End Time of Mute",
    ),
    HeaderFieldDescriptor(
        name="nsamples",
        offset=114,
        format=ScalarType.INT16,
        endianness=Endianness.BIG,
        description="Number of Samples in this Trace",
    ),
    HeaderFieldDescriptor(
        name="sample_interval",
        offset=116,
        format=ScalarType.INT16,
        endianness=Endianness.BIG,
        description="Sample Interval for this Trace",
    ),
    HeaderFieldDescriptor(
        name="gain_type",
        offset=118,
        format=ScalarType.INT16,
        endianness=Endianness.BIG,
        description="Gain Type of Field Instruments",
    ),
    HeaderFieldDescriptor(
        name="instrument_gain",
        offset=120,
        format=ScalarType.INT16,
        endianness=Endianness.BIG,
        description="Instrument Gain Constant",
    ),
    HeaderFieldDescriptor(
        name="instrument_early_gain",
        offset=122,
        format=ScalarType.INT16,
        endianness=Endianness.BIG,
        description="Instrument Early Gain",
    ),
    HeaderFieldDescriptor(
        name="correlated",
        offset=124,
        format=ScalarType.INT16,
        endianness=Endianness.BIG,
        description="Correlated",
    ),
    HeaderFieldDescriptor(
        name="sweep_freq_start",
        offset=126,
        format=ScalarType.INT16,
        endianness=Endianness.BIG,
        description="Sweep Frequency at Start",
    ),
    HeaderFieldDescriptor(
        name="sweep_freq_end",
        offset=128,
        format=ScalarType.INT16,
        endianness=Endianness.BIG,
        description="Sweep Frequency at End",
    ),
    HeaderFieldDescriptor(
        name="sweep_length",
        offset=130,
        format=ScalarType.INT16,
        endianness=Endianness.BIG,
        description="Sweep Length",
    ),
    HeaderFieldDescriptor(
        name="sweep_type",
        offset=132,
        format=ScalarType.INT16,
        endianness=Endianness.BIG,
        description="Sweep Type",
    ),
    HeaderFieldDescriptor(
        name="sweep_trace_taper_start",
        offset=134,
        format=ScalarType.INT16,
        endianness=Endianness.BIG,
        description="Sweep Trace Taper Length at Start",
    ),
    HeaderFieldDescriptor(
        name="sweep_trace_taper_end",
        offset=136,
        format=ScalarType.INT16,
        endianness=Endianness.BIG,
        description="Sweep Trace Taper Length at End",
    ),
    HeaderFieldDescriptor(
        name="taper_type",
        offset=138,
        format=ScalarType.INT16,
        endianness=Endianness.BIG,
        description="Taper Type",
    ),
    HeaderFieldDescriptor(
        name="alias_filter_freq",
        offset=140,
        format=ScalarType.INT16,
        endianness=Endianness.BIG,
        description="Alias Filter Frequency",
    ),
    HeaderFieldDescriptor(
        name="alias_filter_slope",
        offset=142,
        format=ScalarType.INT16,
        endianness=Endianness.BIG,
        description="Alias Filter Slope",
    ),
    HeaderFieldDescriptor(
        name="notch_filter_freq",
        offset=144,
        format=ScalarType.INT16,
        endianness=Endianness.BIG,
        description="Notch Filter Frequency",
    ),
    HeaderFieldDescriptor(
        name="notch_filter_slope",
        offset=146,
        format=ScalarType.INT16,
        endianness=Endianness.BIG,
        description="Notch Filter Slope",
    ),
    HeaderFieldDescriptor(
        name="low_cut_freq",
        offset=148,
        format=ScalarType.INT16,
        endianness=Endianness.BIG,
        description="Low Cut Frequency",
    ),
    HeaderFieldDescriptor(
        name="high_cut_freq",
        offset=150,
        format=ScalarType.INT16,
        endianness=Endianness.BIG,
        description="High Cut Frequency",
    ),
    HeaderFieldDescriptor(
        name="low_cut_slope",
        offset=152,
        format=ScalarType.INT16,
        endianness=Endianness.BIG,
        description="Low Cut Slope",
    ),
    HeaderFieldDescriptor(
        name="high_cut_slope",
        offset=154,
        format=ScalarType.INT16,
        endianness=Endianness.BIG,
        description="High Cut Slope",
    ),
    HeaderFieldDescriptor(
        name="year",
        offset=156,
        format=ScalarType.INT16,
        endianness=Endianness.BIG,
        description="Year Data Recorded",
    ),
    HeaderFieldDescriptor(
        name="day",
        offset=158,
        format=ScalarType.INT16,
        endianness=Endianness.BIG,
        description="Day of Year",
    ),
    HeaderFieldDescriptor(
        name="hour",
        offset=160,
        format=ScalarType.INT16,
        endianness=Endianness.BIG,
        description="Hour of Day",
    ),
    HeaderFieldDescriptor(
        name="minute",
        offset=162,
        format=ScalarType.INT16,
        endianness=Endianness.BIG,
        description="Minute of Hour",
    ),
    HeaderFieldDescriptor(
        name="second",
        offset=164,
        format=ScalarType.INT16,
        endianness=Endianness.BIG,
        description="Second of Minute",
    ),
    HeaderFieldDescriptor(
        name="time_basis_code",
        offset=166,
        format=ScalarType.INT16,
        endianness=Endianness.BIG,
        description="Time Basis Code",
    ),
    HeaderFieldDescriptor(
        name="trace_weighting_factor",
        offset=168,
        format=ScalarType.INT16,
        endianness=Endianness.BIG,
        description="Trace Weighting Factor",
    ),
    HeaderFieldDescriptor(
        name="geophone_group_no_roll1",
        offset=170,
        format=ScalarType.INT16,
        endianness=Endianness.BIG,
        description="Geophone Group Number of Roll Switch Position One",
    ),
    HeaderFieldDescriptor(
        name="geophone_group_no_first_trace",
        offset=172,
        format=ScalarType.INT16,
        endianness=Endianness.BIG,
        description="Geophone Group Number of Trace Number One within Original Field Record",  # noqa: E501
    ),
    HeaderFieldDescriptor(
        name="geophone_group_no_last_trace",
        offset=174,
        format=ScalarType.INT16,
        endianness=Endianness.BIG,
        description="Geophone Group Number of Last Trace within Original Field Record",  # noqa: E501
    ),
    HeaderFieldDescriptor(
        name="gap_size",
        offset=176,
        format=ScalarType.INT16,
        endianness=Endianness.BIG,
        description="Gap Size (total number of groups dropped)",
    ),
    HeaderFieldDescriptor(
        name="over_travel",
        offset=178,
        format=ScalarType.INT16,
        endianness=Endianness.BIG,
        description="Over Travel Associated with Taper",
    ),
]


rev0_textual_file_header = TextHeaderDescriptor(
    rows=40,
    cols=80,
    offset=0,
    encoding=TextHeaderEncoding.EBCDIC,
    format=ScalarType.UINT8,  # noqa: A003
)


rev0_binary_file_header = BinaryHeaderDescriptor(
    fields=BINARY_FILE_HEADER_FIELDS_REV0,
    item_size=400,
    offset=3200,
)


rev0_trace_header = TraceHeaderDescriptor(
    fields=TRACE_HEADER_FIELDS_REV0,
    item_size=240,
)


rev0_trace_data = TraceDataDescriptor(
    format=ScalarType.IBM32,  # noqa: A003
    endianness=Endianness.BIG,
)


rev0_trace = TraceDescriptor(
    header_descriptor=rev0_trace_header,
    data_descriptor=rev0_trace_data,
)


rev0_segy = SegyDescriptor(
    segy_standard=SegyStandard.REV0,
    text_file_header=rev0_textual_file_header,
    binary_file_header=rev0_binary_file_header,
    trace=rev0_trace,
)



"""Core functionality for SEG-Y ninja templates."""
from __future__ import annotations

from abc import abstractmethod
from typing import TYPE_CHECKING

from pydantic import BaseModel
from pydantic import ConfigDict
from pydantic import Field
from pydantic.alias_generators import to_camel

if TYPE_CHECKING:
    from typing import Any

    import numpy as np


class CamelCaseModel(BaseModel):
    """Base model with camel case aliases. Extends BaseModel.

    Attributes:
        model_config: The configuration dictionary for the model.
            - alias_generator: The function used to generate aliases.
            - populate_by_name: Flag indicating whether to populate by name.
    """

    model_config = ConfigDict(alias_generator=to_camel, populate_by_name=True)

    def model_dump(self, *args: Any, **kwargs: Any) -> dict[str, Any]:  # noqa: ANN401
        """Dump the model into a dictionary by alias."""
        return super().model_dump(*args, **kwargs, by_alias=True)

    def model_dump_json(self, *args: Any, **kwargs: Any) -> str:  # noqa: ANN401
        """Dump the model into a JSON string by alias."""
        return super().model_dump_json(*args, **kwargs, by_alias=True)


class BaseTypeDescriptor(CamelCaseModel):
    """A base model for all SEG-Y Ninja types."""

    description: str | None = Field(
        default=None, description="Description of the field."
    )

    @property
    @abstractmethod
    def dtype(self) -> np.dtype[Any]:
        """Abstract property for subclasses to return the numpy dtype."""

    @property
    def itemsize(self) -> int:
        """Number of bytes for the data type."""
        return self.dtype.itemsize



"""Data models representing data types."""


from __future__ import annotations

from typing import Any

import numpy as np
from pydantic import Field

from segy.compat import StrEnum
from segy.schema.base import BaseTypeDescriptor


class Endianness(StrEnum):
    """Enum class representing endianness."""

    BIG = "big"
    LITTLE = "little"

    @property
    def symbol(self) -> str:
        """Get the numpy symbol for the endianness."""
        return ">" if self == Endianness.BIG else "<"


class ScalarType(StrEnum):
    """A class representing different data types used for data formatting."""

    IBM32 = "ibm32"
    INT64 = "int64"
    INT32 = "int32"
    INT16 = "int16"
    INT8 = "int8"
    UINT64 = "uint64"
    UINT32 = "uint32"
    UINT16 = "uint16"
    UINT8 = "uint8"
    FLOAT64 = "float64"
    FLOAT32 = "float32"
    FLOAT16 = "float16"

    @property
    def char(self) -> str:
        """Returns the numpy character code for a given data type string."""
        if self.value == "ibm32":
            return np.sctype2char("uint32")  # noqa: NPY201

        return np.sctype2char(str(self.value))  # noqa: NPY201


class DataTypeDescriptor(BaseTypeDescriptor):
    """A descriptor class for a scalar data type with endianness support."""

    format: ScalarType = Field(..., description="The data type of the field.")  # noqa: A003
    endianness: Endianness = Field(
        default=Endianness.BIG, description="The byte order of the field."
    )

    @property
    def dtype(self) -> np.dtype[Any]:
        """Converts the byte order and data type of the object into a NumPy dtype."""
        symbol = self.endianness.symbol
        char = self.format.char

        return np.dtype(symbol + char)


class StructuredFieldDescriptor(DataTypeDescriptor):
    """A descriptor class for a structured data-type field."""

    name: str = Field(..., description="The short name of the field.")
    offset: int = Field(..., ge=0, description="Starting byte offset.")


class StructuredDataTypeDescriptor(BaseTypeDescriptor):
    """A descriptor class for a structured array data-type."""

    fields: list[StructuredFieldDescriptor] = Field(
        ..., description="Fields of the structured data type."
    )
    item_size: int | None = Field(
        default=None, description="Expected size of the struct."
    )
    offset: int | None = Field(default=None, ge=0, description="Starting byte offset.")

    @property
    def dtype(self) -> np.dtype[Any]:
        """Get numpy dtype."""
        names = [field.name for field in self.fields]
        offsets = [field.offset for field in self.fields]
        formats = [field.dtype for field in self.fields]

        dtype_conf = {
            "names": names,
            "formats": formats,
            "offsets": offsets,
        }

        if self.item_size is not None:
            dtype_conf["itemsize"] = self.item_size

        return np.dtype(dtype_conf)  # type: ignore



"""Common components for the CLI."""


from __future__ import annotations

from pathlib import Path
from typing import Annotated
from typing import Optional

import typer

UriArgument = Annotated[
    str, typer.Argument(help="Valid URI for loading the SEG-Y file.")
]

ListOfIntegersOption = Annotated[list[int], typer.Option(help="List of integers.")]

ListOfFieldNamesOption = Annotated[
    list[str], typer.Option(default_factory=list, help="List of field names.")
]

JsonFileOutOption = Annotated[
    Optional[Path], typer.Option(help="Path for JSON output.")
]

TextFileOutOption = Annotated[
    Optional[Path], typer.Option(help="Path for text output.")
]


def modify_path(
    path: Path, suffix: str, default_extension: str, delimiter: str = "_"
) -> Path:
    """Modify a path with a suffix appended and ensure default extension is honored."""
    new_stem = f"{path.stem}{delimiter}{suffix}"

    if path.suffix:  # If there's an existing extension
        extension = path.suffix
        extension = default_extension if extension != default_extension else extension
        new_name = f"{new_stem}{extension}"
    else:  # If there's no extension
        new_name = f"{new_stem}{default_extension}"

    return path.with_name(new_name)
